C THIS IS THE FORTRAN66 VERSION OF FSUDOKU
C THE FOLLOWING KINDS ARE SUPPORTED:
C STD:  THE CLASSIC SUDOKU
C WIN:  THE SUDOKU AS PUBLISHED BY THE NRC: 4 EXTRA UNITS OF 3X3,
C       LOCATED WITH ONE ROW/COLUMN SPACING FROM THE OUTER EDGE
C       ALSO KNOWN AS WINDOKU. SEE HTTPS://WWW.SUDOKUWIKI.ORG/WINDOKU.ASPX
C XXX:  AKA SUDOKU-X: THE DIAGONALS ALSO PLAY A ROLE

C INPUT IS THE TYPE (WIN OR STD OR XXX AND THE BOARD, FOR EXAMPLE (WITHHOUT THE
C LEADING 'C '):

C STD    THIS IS A SIMPLE SUDOKU
C 0 0 0   0 0 0   7 2 0
C 0 1 0   5 3 0   0 6 4
C 6 0 0   0 0 2   1 0 0
C 3 0 0   9 0 0   0 4 0
C 0 0 0   8 0 4   0 0 0
C 0 5 0   0 0 1   0 0 2
C 0 0 6   4 0 0   0 0 8
C 5 8 0   0 9 7   0 3 0
C 0 2 3   0 0 0   0 0 0

C A POINT ('.') IS INTERPRETED AS 0

      CALL SUDOKU
      END

C    THE MAIN SUBROUTINE

      SUBROUTINE SUDOKU
         COMMON /GAME/ TYPE, NUMSOL, WIN, STD, XXX,
     +                 A(9,9), AA(9,9), UNIT(4,31), INUNIT(4,9,9)
         INTEGER A, UNIT, INUNIT, TYPE, NUMSOL, WIN, STD, XXX
         LOGICAL AA

         INTEGER NMAX
         LOGICAL CORREC

         NMAX = 100
         NUMSOL = 0
         WRITE(6,1000)
 1000    FORMAT(6HINPUT:)
         CALL READB
         CALL INIT
         CALL PRINTB
         IF (CORREC()) GOTO 1020
         WRITE(6,1010)
 1010    FORMAT('THIS IS NOT A VALID SUDOKU')
         STOP
 1020    CONTINUE
         CALL SOLVER(NMAX)
         WRITE(6,1030) NUMSOL
 1030    FORMAT(20HNUMBER OF SOLUTIONS:,I10)
         IF (NUMSOL .LT. NMAX) GOTO 1050
         WRITE(6,1040)
 1040    FORMAT(34HPROBABLY, THERE ARE MORE SOLUTIONS)
 1050    CONTINUE
      END

C    READ THE BOARD

      SUBROUTINE READB
         COMMON /GAME/ TYPE, NUMSOL, WIN, STD, XXX,
     +                 A(9,9), AA(9,9), UNIT(4,31), INUNIT(4,9,9)
         INTEGER A, UNIT, INUNIT, TYPE, NUMSOL, WIN, STD, XXX
         LOGICAL AA

         INTEGER NUMS(10),SPACE,POINT
         DATA NUMS/1H0,1H1,1H2,1H3,1H4,1H5,1H6,1H7,1H8,1H9/
         DATA SPACE/1H /,POINT/1H./
         INTEGER LINE(200)
         INTEGER I,J,K,L

         READ (5,1060) TYPE
 1060    FORMAT(A3)
         IF (TYPE .NE. XXX .AND. TYPE .NE. WIN) TYPE = STD

         DO 1070 I=1,9
            DO 1070 J=1,9
 1070    A(I,J) = -1
         I = 1
         J = 1
 1080    CONTINUE
         DO 1090 K=1,200
 1090    LINE(K) = SPACE
         READ(5,1100,END=1140) (LINE(K),K=1,200)
 1100    FORMAT(200A1)
         DO 1130 K=1,200
            IF (LINE(K) .EQ. POINT) LINE(K) = NUMS(1)
            DO 1120 L=1,10
               IF (LINE(K) - NUMS(L)) 1120,1110,1120
 1110          A(I,J) = L - 1
               J = J+1
               IF (J .LE. 9) GOTO 1130
               I = I+1
               IF (I .GT. 9) RETURN
               J = 1
 1120       CONTINUE
 1130    CONTINUE
         IF (I .LE. 9) GOTO 1080
         RETURN
 1140    CONTINUE
         WRITE(6,1150)I,J
 1150    FORMAT(15HCANNOT READ ROW,I3,7H COLUMN,I3)
         STOP
      END

C    PRINT THE BOARD
      SUBROUTINE PRINTB
         COMMON /GAME/ TYPE, NUMSOL, WIN, STD, XXX,
     +                 A(9,9), AA(9,9), UNIT(4,31), INUNIT(4,9,9)
         INTEGER A, UNIT, INUNIT, TYPE, NUMSOL, WIN, STD, XXX
         LOGICAL AA

         INTEGER I,J
         WRITE(6,1160) TYPE
 1160    FORMAT(A3)
         DO 1170 I=1,9
            WRITE(6,1180)(A(I,J),J=1,9)
            IF (MOD(I,3) .EQ. 0) WRITE(6,1190)
 1170    CONTINUE

 1180    FORMAT(3(3I2,1H ,3I2,1H ,3I2))
 1190    FORMAT(1H )

      END

C    INITIALIZE THINGS

      SUBROUTINE INIT

C UNIT WILL CONTAIN START AND END COORDINATES OF UNITS
C UNIT(1,:) STARTROW
C UNIT(2,:) ENDROW
C UNIT(3,:) STARTCOLUMN
C UNIT(4,:) ENDCOLUMN

         COMMON /GAME/ TYPE, NUMSOL, WIN, STD, XXX,
     +                 A(9,9), AA(9,9), UNIT(4,31), INUNIT(4,9,9)
         INTEGER A, UNIT, INUNIT, TYPE, NUMSOL, WIN, STD, XXX
         LOGICAL AA

         INTEGER K,I,J,P,KM
         K=1
C     ROWS
         DO 1200 I=1,9
            UNIT(1,K) = I
            UNIT(2,K) = I
            UNIT(3,K) = 1
            UNIT(4,K) = 9
            K = K+1
 1200    CONTINUE
C     COLUMNS
         DO 1210 I=1,9
            UNIT(1,K) = 1
            UNIT(2,K) = 9
            UNIT(3,K) = I
            UNIT(4,K) = I
            K = K+1
 1210    CONTINUE
C     SQUARES
         DO 1230 I=1,3
            DO 1220 J=1,3
               UNIT(1,K) = (I-1)*3 + 1
               UNIT(2,K) = UNIT(1,K) + 2
               UNIT(3,K) = (J-1)*3 + 1
               UNIT(4,K) = UNIT(3,K) + 2
               K = K+1
 1220       CONTINUE
 1230    CONTINUE
         IF (TYPE .NE. WIN) GOTO 1260
C     WINDOKU SQUARES
         DO 1250 I=1,2
            DO 1240 J=1,2
               UNIT(1,K) = (I-1)*4 + 2
               UNIT(2,K) = UNIT(1,K) + 2
               UNIT(3,K) = (J-1)*4 + 2
               UNIT(4,K) = UNIT(3,K) + 2
               K = K+1
 1240       CONTINUE
 1250    CONTINUE
 1260    CONTINUE
         KM = K

         DO 1270 K = 1,4
            DO 1270 I=1,9
               DO 1270 J=1,9
                  INUNIT(K,I,J) = 0
 1270    CONTINUE

         DO 1300 I=1,9
            DO 1300 J=1,9
               P = 0
               DO 1290 K=1,KM
                  IF (I .GE. UNIT(1,K) .AND.
     +             I .LE. UNIT(2,K) .AND.
     +             J .GE. UNIT(3,K) .AND.
     +             J .LE. UNIT(4,K) ) GOTO 1280
                  GOTO 1290
 1280             P = P+1
                  INUNIT(P,I,J) = K
 1290          CONTINUE
 1300    CONTINUE

      END

C CHECK IF A(R(1):R(2),R(3):R(4)) CONTAINS N

      LOGICAL FUNCTION HAS(R,N)
         INTEGER R(4)
         INTEGER N
         COMMON /GAME/ TYPE, NUMSOL, WIN, STD, XXX,
     +                 A(9,9), AA(9,9), UNIT(4,31), INUNIT(4,9,9)
         INTEGER A, UNIT, INUNIT, TYPE, NUMSOL, WIN, STD, XXX
         LOGICAL AA

         INTEGER I,J
         HAS = .TRUE.
         DO 1310 I=R(1),R(2)
            DO 1310 J=R(3),R(4)
               IF (A(I,J) .EQ. N) RETURN
 1310    CONTINUE
         HAS = .FALSE.
      END

C CHECK IF A(I,J) IS POSSIBLE

      LOGICAL FUNCTION POSSIB(I,J)
         INTEGER I, J
         COMMON /GAME/ TYPE, NUMSOL, WIN, STD, XXX,
     +                 A(9,9), AA(9,9), UNIT(4,31), INUNIT(4,9,9)
         INTEGER A, UNIT, INUNIT, TYPE, NUMSOL, WIN, STD, XXX
         LOGICAL AA

         INTEGER K, II, AIJ
         LOGICAL HAS

         POSSIB = .FALSE.

         AIJ    = A(I,J)
         A(I,J) = 0

C    3X3 BLOCKS AND ROWS AND COLUMNS
         DO 1330 K=1,4
            II = INUNIT(K,I,J)
            IF (II) 1320,1340,1320
 1320       IF (HAS(UNIT(1,II),AIJ)) GOTO 1440
 1330    CONTINUE
 1340    CONTINUE

C    DIAGONALS:
         IF(TYPE - XXX) 1430,1350,1430
 1350    IF(I-J) 1390,1360,1390
 1360    DO 1380 II=1,9
            IF(AIJ - A(II,II)) 1380,1370,1380
 1370       GOTO 1440
 1380    CONTINUE
 1390    CONTINUE
         IF (I-10+J) 1430,1400,1430
 1400    DO 1420 II=1,9
            IF(AIJ-A(II,10-II)) 1420,1410,1420
 1410       GOTO 1440
 1420    CONTINUE
 1430    CONTINUE
         POSSIB = .TRUE.
 1440    A(I,J) = AIJ
         RETURN
      END

C RETURNS .TRUE. IF THIS IS A VALID SUDOKU

      LOGICAL FUNCTION CORREC()
         COMMON /GAME/ TYPE, NUMSOL, WIN, STD, XXX,
     +                 A(9,9), AA(9,9), UNIT(4,31), INUNIT(4,9,9)
         INTEGER A, UNIT, INUNIT, TYPE, NUMSOL, WIN, STD, XXX
         LOGICAL AA

         INTEGER I,J
         LOGICAL POSSIB
         CORREC = .TRUE.
         DO 1490 I=1,9
            DO 1480 J=1,9
               IF(A(I,J) .GE. 0 .AND. A(I,J) .LE. 9)  GOTO 1450
               CORREC = .FALSE.
               RETURN
 1450          CONTINUE
               IF(A(I,J)) 1460,1470,1460
 1460          CONTINUE
               IF (POSSIB(I,J)) GOTO 1470
               CORREC = .FALSE.
               RETURN
 1470          CONTINUE
 1480       CONTINUE
 1490    CONTINUE
      END

C    RETURNS I,J OF NEXT CELL THAT HAS TO BE DETERMINED
      SUBROUTINE NEXTC(I,J)
         INTEGER I,J
         COMMON /GAME/ TYPE, NUMSOL, WIN, STD, XXX,
     +                 A(9,9), AA(9,9),UNIT(4,31), INUNIT(4,9,9)
         INTEGER A, UNIT, INUNIT, TYPE, NUMSOL, WIN, STD, XXX
         LOGICAL AA

 1500    J = J+1
         IF (J .LE. 9) GOTO 1510
         I = I+1
         J = 1
 1510    IF (I .GT. 9) RETURN
         IF (.NOT. AA(I,J)) RETURN
         GOTO 1500
      END

C    RETURNS I,J OF PREVIOUS CELL THAT HAS TO BE DETERMINED
      SUBROUTINE PREVC(I,J)
         INTEGER I,J
         COMMON /GAME/ TYPE, NUMSOL, WIN, STD, XXX,
     +                 A(9,9), AA(9,9),UNIT(4,31), INUNIT(4,9,9)
         INTEGER A, UNIT, INUNIT, TYPE, NUMSOL, WIN, STD, XXX
         LOGICAL AA

 1520    J = J-1
         IF (J .GE. 1) GOTO 1530
         I = I-1
         J = 9
 1530    IF (I .LT. 1) RETURN
         IF (.NOT. AA(I,J)) RETURN
         GOTO 1520
      END

      SUBROUTINE HANSOL(N)
         INTEGER N
         COMMON /GAME/ TYPE, NUMSOL, WIN, STD, XXX,
     +                 A(9,9), AA(9,9),UNIT(4,31), INUNIT(4,9,9)
         INTEGER A, UNIT, INUNIT, TYPE, NUMSOL, WIN, STD, XXX
         LOGICAL AA

         NUMSOL = NUMSOL + 1
         WRITE(6,1540) NUMSOL
 1540    FORMAT('SOLUTION NR ',I8)
         CALL PRINTB
         IF (NUMSOL .GE. N) CALL EPILOG(N)
      END


      SUBROUTINE EPILOG(N)
         INTEGER N
         COMMON /GAME/ TYPE, NUMSOL, WIN, STD, XXX,
     +                 A(9,9), AA(9,9),UNIT(4,31), INUNIT(4,9,9)
         INTEGER A, UNIT, INUNIT, TYPE, NUMSOL, WIN, STD, XXX
         LOGICAL AA

         WRITE(6,1550) NUMSOL
 1550    FORMAT('NUMBER OF SOLUTIONS IS',I8)
         IF (NUMSOL .GE. N) WRITE(6,1560)
 1560    FORMAT('PROBABLY, THERE ARE MORE SOLUTIONS')
         STOP
      END

C SOLVES SUDOKU, AT MOST NMAX SOLUTIONS ARE PRINTED
      SUBROUTINE SOLVER(NMAX)
         INTEGER NMAX
         COMMON /GAME/ TYPE, NUMSOL, WIN, STD, XXX,
     +                 A(9,9), AA(9,9),UNIT(4,31), INUNIT(4,9,9)
         INTEGER A, UNIT, INUNIT, TYPE, NUMSOL, WIN, STD, XXX
         LOGICAL AA

         LOGICAL POSSIB
         INTEGER I,J

         DO 1570 I=1,9
            DO 1570 J=1,9
 1570    AA(I,J) = A(I,J) .NE. 0

         I = 0
         J = 9

C     NEXTC WILL POINT (I,J) TO THE FIRST MODIFIABLE CELL
         CALL NEXTC(I,J)

C     IF (I,J) IS OUTSIDE BOARD, WE HAVE AN ALREADY SOLVED SUDOKU:
         IF (I .LE. 9) GOTO 1580
         CALL HANSOL(NMAX)
         RETURN

 1580    CONTINUE

C     INVARIANT: AT THIS POINT, I,J IS ALWAYS THE CELL UNDER INVESTIGATION

C    TRY THE NEXT VALUE FOR A(I,J):
         A(I,J) = A(I,J) + 1
         IF (A(I,J) .LE. 9) GOTO 1590
C    NO SOLUTION GIVEN THE PREVIOUS CELLS: CLEAR AND STEP BACK
         A(I,J) = 0
         CALL PREVC(I,J)
C    IF WE LEAVE THE BOARD, THERE IS NOTHING MORE TO INVESTIGATE
         IF (I .LT. 1) GOTO 1600
         GOTO 1580

 1590    CONTINUE

C     IF A(I,J) HAS A POSSIBLE VALUE, GO TO THE NEXT CELL
         IF (.NOT. POSSIB(I,J)) GOTO 1580
         CALL NEXTC(I,J)
         IF (I .LE. 9) GOTO 1580
C     STEPPED OUTSIDE BOARD, SO WE HAVE A SOLUTION
         CALL HANSOL(NMAX)
C     PREVC WILL LET (I,J) POINT TO THE LAST MODIFIABLE CELL
         CALL PREVC(I,J)
         GOTO 1580

 1600    CONTINUE

      END

      BLOCK DATA
         COMMON /GAME/ TYPE, NUMSOL, WIN, STD, XXX,
     +                 A(9,9), AA(9,9), UNIT(4,31), INUNIT(4,9,9)
         INTEGER A, UNIT, INUNIT, TYPE, NUMSOL, WIN, STD, XXX
         LOGICAL AA
         DATA WIN/3HWIN/, STD/3HSTD/, XXX/3HXXX/
      END

